/**
 * SWE KEx-DatexII (de.bsvrz.kex.datex2)
 * 
 * Copyright (C) 2014 TraffGo Road GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * TraffGo Road GmbH
 * Ostwall 11-19
 * 47798 Krefeld
 * Phone: +49-2151-3295058
 * mailto: info@traffgoroad.de
 */

package de.bsvrz.kex.datex2.allgemein.impl;

import java.io.NotActiveException;
import java.io.Serializable;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.concurrent.ConcurrentHashMap;

import javax.activation.UnsupportedDataTypeException;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.ClientReceiverInterface;
import de.bsvrz.dav.daf.main.ClientSenderInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.Data.Array;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.ReceiveOptions;
import de.bsvrz.dav.daf.main.ReceiverRole;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.SenderRole;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.kex.datex2.allgemein.KonfigurationsModul;
import de.bsvrz.kex.datex2.allgemein.PersistenzModul;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Implementierung des Interface PersistenzModul; Daten werden ber den
 * Datenverteiler in einer speziellen Attributgruppe persistiert.
 * 
 * @author TraffGo Road GmbH
 * @version $Id$
 */
public class PersistenzUeberDaV implements PersistenzModul, ClientReceiverInterface, ClientSenderInterface {

  /** Datenverteiler-Feldname Wert.*/
  private static final String ATT_VALUE = "Wert";

  /** Datenverteiler-Feldname Schlssel. */
  private static final String ATT_KEY = "Schlssel";

  /** Datenverteiler-Feldname Schlssel-Wert-Eintrag. */
  private static final String ATL_KEY_VALUE_ENTRY = "Schlssel-Wert-Eintrag";

  /** Aspekt zum Schreiben von Parametern. */
  private static final String ASP_SCHREIBE_PARAM = "asp.parameterVorgabe";

  /** Aspekt zum Lesen von Parametern. */
  private static final String ASP_LESE_PARAM = "asp.parameterSoll";

  /** Attributgruppe der Wertetabelle. */
  private static final String ATG_WERTETABELLE = "atg.werteTabelle";

  /** Schlssel fr Angabe des Datenverteiler-Objekts, an welches die Persistenz gehngt wird. */
  private static final String KEY_PERSISTENZ_OBJEKT = "PersistenzObjekt";

  /** Maximale Wartezeit auf positive Sendesteuerung (in Millisekunden). */
  private static final long MAX_WAIT = 5000;
  
  /** Datenverteiler-Verbindung. */
  private ClientDavInterface davConnection;
  
  /** Referenz auf Konfigurationsmodul. */
  private KonfigurationsModul config;
  
  /** Prfix fr Log-Ausgaben. */
  private String logPrefix = "";
  
  /** Speichert korrekte Initialisierung. */
  private boolean isInitialized = false;
  
  /** Datenverteiler-Objekt, an dem die Persistenz hngt. */
  private SystemObject persistenzObjekt;
  
  /** Lokale Speicherung der Wertetabelle. */  
  private Map<String, String> keyValueStore = new ConcurrentHashMap<>();
  
  /** Logging. */
  private static final Debug LOGGER = Debug.getLogger();
  
  /**
   * Konstruktor.
   * 
   * @param davConnection Datenverteiler-Verbindung
   * @param config Referenz auf Konfigurationsmodul
   */
  public PersistenzUeberDaV(final ClientDavInterface davConnection, final KonfigurationsModul config) {
    this.davConnection = davConnection;
    this.config = config;
    this.logPrefix = "In Kontext '" + config.getContext() + "': ";
    try {
      String persistenzPid = this.config.getString(KEY_PERSISTENZ_OBJEKT);
      DataDescription dd = new DataDescription(davConnection.getDataModel().getAttributeGroup(ATG_WERTETABELLE), 
          davConnection.getDataModel().getAspect(ASP_LESE_PARAM));
      persistenzObjekt = davConnection.getDataModel().getObject(persistenzPid);    
      davConnection.subscribeReceiver(this, persistenzObjekt,  dd,  ReceiveOptions.normal(),  ReceiverRole.receiver());

      DataDescription ddSend = new DataDescription(davConnection.getDataModel().getAttributeGroup(ATG_WERTETABELLE), 
          davConnection.getDataModel().getAspect(ASP_SCHREIBE_PARAM));

      davConnection.subscribeSender(this, persistenzObjekt,  ddSend,  SenderRole.sender());
    } catch (Exception e) {
      LOGGER.error(logPrefix + "Persistenz-Schicht konnte nicht initialisiert werden.", e);
    }    
  }
  
  @Override
  public String getContext() {
    return persistenzObjekt.getPid();
  }

  @Override
  public <T extends Serializable> T getData(final String key, final Class<T> classx)
      throws UnsupportedDataTypeException, NotActiveException {
    if (!isInitialized) {
      try {
        Thread.sleep(MAX_WAIT);
      } catch (InterruptedException dummy) {
        throw new NotActiveException("Warten auf Antwort abgebrochen.");
      }
      if (!isInitialized) {
        throw new NotActiveException("Keine Antwort vom Datenverteiler.");
      }
    }
    String value = keyValueStore.get(key);
    if (value == null) {
      return null;
    }
    if (classx.equals(String.class)) {
      return classx.cast(value);
    }
    if (classx.equals(Integer.class)) {
      return classx.cast(new Integer(value));
    }
    if (classx.equals(Float.class)) {
      return classx.cast(new Float(value));
    }
    throw new UnsupportedDataTypeException();
  }

  @Override
  public String getString(final String key) {
    try {
      return getData(key, String.class);
    } catch (UnsupportedDataTypeException ex) {
      throw new Error(); // this indicates a serious problem with the JVM
    } catch (NotActiveException e) {
      return null;
    }
  }

  @Override
  public Integer getInteger(final String key) {
    try {
      return getData(key, Integer.class);
    } catch (UnsupportedDataTypeException ex) {
      throw new Error();
    } catch (NotActiveException e) {
      return null;
    }
  }

  @Override
  public <T extends Serializable> void putData(final String key, final T value) {
    keyValueStore.put(key, value.toString());

    if (persistenzObjekt != null) {
      DataDescription dd = new DataDescription(davConnection.getDataModel().getAttributeGroup(ATG_WERTETABELLE), 
          davConnection.getDataModel().getAspect(ASP_SCHREIBE_PARAM));

      AttributeGroup atg = davConnection.getDataModel().getAttributeGroup(ATG_WERTETABELLE);

      Data data = davConnection.createData(atg);

      try {
        Array ar = data.getItem(ATL_KEY_VALUE_ENTRY).asArray();
        ar.setLength(keyValueStore.size());

        int idx = 0;
        for (String ikey: keyValueStore.keySet()) {
          ar.getItem(idx).getItem(ATT_KEY).asTextValue().setText(ikey);
          ar.getItem(idx).getItem(ATT_VALUE).asTextValue().setText(keyValueStore.get(ikey));
          idx++;
        }
      } catch (NoSuchElementException ex) {
        LOGGER.error("Kann nicht persistieren, Datenmodell passt nicht zur Software.", ex);
      }
 
      LOGGER.fine(logPrefix + "Schreibe Persistenzdaten: " + data);

      try {
        davConnection.sendData(new ResultData(persistenzObjekt, dd, davConnection.getTime(), data));
      } catch (Exception e) {
        LOGGER.error(logPrefix + "Persistierung nicht mglich.", e);
      }
    }
  }
  
  @Override
  public boolean hasData(final String key) {
    return keyValueStore.get(key) != null;
  }

  @Override
  public void update(final ResultData[] resultDataArray) {
    isInitialized = true;
    
    LOGGER.fine(logPrefix + "Update der Persistenzdaten (" + resultDataArray.length + ")");
    
    for (ResultData data : resultDataArray) {
      if ((data != null) && (data.getData() != null)) {
        LOGGER.fine("Empfangen: " + data.toString());
        Array ar = data.getData().getItem(ATL_KEY_VALUE_ENTRY).asArray();
        for (int i = 0; i < ar.getLength(); i++) {
          try {
            String key = ar.getItem(i).getItem(ATT_KEY).asTextValue().getText();
            String value = ar.getItem(i).getItem(ATT_VALUE).asTextValue().getText();
      
            keyValueStore.put(key, value);
          } catch (NullPointerException npe) {
            npe.printStackTrace();
          }
        }
      }
    }
  }

  @Override
  public void dataRequest(final SystemObject arg0, final DataDescription arg1, final byte arg2) {
  }

  @Override
  public boolean isRequestSupported(final SystemObject arg0, final DataDescription arg1) {
    return false;
  }
}
