/**
 * SWE KEx-DatexII (de.bsvrz.kex.datex2)
 * 
 * Copyright (C) 2014 TraffGo Road GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * TraffGo Road GmbH
 * Ostwall 11-19
 * 47798 Krefeld
 * Phone: +49-2151-3295058
 * mailto: info@traffgoroad.de
 */

package de.bsvrz.kex.datex2.plugins.ausfuehrung.impl;

import java.io.ByteArrayOutputStream;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Collection;
import java.util.Date;
import java.util.Map;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Marshaller;
import javax.xml.namespace.QName;

import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.kex.datex2.allgemein.ObjekteAtgAsp;
import de.bsvrz.kex.datex2.plugins.ausfuehrung.PluginAusfuehrungsAlgorithmus;
import de.bsvrz.kex.datex2.plugins.datensammlung.DatenSammlungsMethode;
import de.bsvrz.kex.datex2.plugins.definition.PluginExport;
import de.bsvrz.kex.datex2.plugins.konvertierung.DatexKonvertierungStatisch;
import de.bsvrz.kex.datex2.plugins.konvertierung.DatexKonvertierungStatisch.CreationOptions;
import de.bsvrz.kex.datex2.plugins.publizierung.PublizierungsMethode;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Klasse implementiert einen Export-Ausfhrungsalgorithmus fr Plugins.
 * 
 * @author TraffGo Road GmbH
 * @version $Id$
 */

public class StandardExportPluginAusfuehrung implements PluginAusfuehrungsAlgorithmus {
  /** Schlssel fr Speicherung von Hash-Werten. */
  private static final String HASH_CODE_KEY = "HashCodeStaticXML";
  
  /** Zeichencodierung. */
  private static final String CHAR_ENCODING = "UTF-8";

  /** Datex II Namensraum. */
  private static final String DATEX2_NAMESPACE = "http://datex2.eu/schema/2/2_0";
  
  /** Root-Tag des Datex II Formats. */
  private static final String DATEX2_ROOTTAG = "d2LogicalModel";

  /** Referenz auf Plugin-Definitions-Klasse. */
  private PluginExport plugin;
  
  /** Prfix fr Log-Ausgaben. */
  private String logPrefix = "";

  /** Logging. */
  private static final Debug LOGGER = Debug.getLogger();
  
  /**
   * Konstruktor.
   * 
   * @param plugin Referenz auf Plugin-Definitions-Klasse
   */
  
  public StandardExportPluginAusfuehrung(final PluginExport plugin) {
    this.plugin = plugin;
  }

  @Override
  public boolean initialisiere() {
    boolean hasConfig = true;
    String context = plugin.getKonfigurationsModul().getContext();
    logPrefix = "In Kontext '" + context + "': ";
//    for (PublizierungsMethode m : plugin.getStatischePublizierungsMethoden()) {  
//      String keyStatic = m.getPublizierungsMethodenName() + "-static-URL";
//      
//      if (!plugin.getKonfigurationsModul().hasData(keyStatic)) {
//        hasConfig = false;
//        LOGGER.error("In Kontext '"+context+ "': Fehlender Schlssel '" + keyStatic + "'.");
//      }
//    }
//    for (PublizierungsMethode m : plugin.getDynamischePublizierungsMethoden()) {    
//      String keyDynamic = m.getPublizierungsMethodenName() + "-dynamic-URL";
//      if (!plugin.getKonfigurationsModul().hasData(keyDynamic)) {
//        hasConfig = false;
//        LOGGER.error("In Kontext '"+context+ "': Fehlender Schlssel '" + keyDynamic + "'.");
//      }
//    }
    return hasConfig;
  }

  @Override
  public void starte() {
    try {
      Collection<ObjekteAtgAsp> staticObjects = plugin.getObjects();

      for (ObjekteAtgAsp o : staticObjects) {
        o.sortiereObjekte();
      }

      DatexKonvertierungStatisch pstatic = plugin.getDatexKonvertierungStatisch();
      if (pstatic != null) {
        Object xmlStaticObjectsSorted = pstatic.getDatexII(staticObjects, CreationOptions.VERSION_FREE);

        if (xmlStaticObjectsSorted == null) {
          throw new Exception("Leeres XML-Dokument als Ergebnis.");
        } else {
          Integer hashCode = bestimmeHashCode(xmlStaticObjectsSorted, pstatic.getJAXBContext());
          
//          Integer hashCode = xmlStaticObjectsSorted.toString().replace("<publicationTime>.*</publicationTime>","").hashCode();
          Integer storedHashCode = null;
          if (plugin.getPersistenzModul() != null) {
            try {
              storedHashCode = plugin.getPersistenzModul().getInteger(HASH_CODE_KEY);
            } catch (Exception ex) {
              LOGGER.warning(logPrefix 
                  + "Zugriff auf Persistenz-Schicht fhrte zu einem Fehler, statische Daten werden nicht publiziert.");
            }
          }

          LOGGER.info(logPrefix + "Hash-Code neu/alt: " + hashCode + " / " + storedHashCode);

          if ((storedHashCode != null) && (!hashCode.equals(storedHashCode))) {
            xmlStaticObjectsSorted = pstatic.getDatexII(staticObjects);

            for (PublizierungsMethode m : plugin.getStatischePublizierungsMethoden()) {
              LOGGER.info(logPrefix + "Beginn der Publikation statischer Daten fr " + m.getPublizierungsMethodenName());
              String destination = plugin.getKonfigurationsModul().getString(m.getPublizierungsMethodenName() + "-static-URL");
              m.publiziere(xmlStaticObjectsSorted, plugin.getDatexKonvertierungStatisch().getJAXBContext(), destination);
              LOGGER.info(logPrefix + "Statische Daten publiziert.");
            }
            if (plugin.getPersistenzModul() != null) {
              plugin.getPersistenzModul().putData(HASH_CODE_KEY, hashCode);
            } else {
              LOGGER.warning(logPrefix + "Kein Persistenz-Modul aktiv. Statische Daten werden bei jedem Start publiziert.");
            }
          } else {
            LOGGER.info(logPrefix + "berspringe Publikation statischer Daten, da offenbar keine nderungen.");
          }
        }
      }

      DatenSammlungsMethode dataMethod = plugin.getDatenSammlungsMethode();        
      dataMethod.initialisiere(staticObjects);
    } catch (Exception ex) {
      LOGGER.error(logPrefix + "Publikation statischer Daten konnte wegen eines Fehlers nicht durchgefhrt werden." , ex);   
    }
  }

  /**
   * Ermittelt den Hashcode der XML-Struktur (aus dem resultierenden Text bestimmt).
   * 
   * @param xml JAXB Objekt
   * @param context JAXB-Context
   * @return Integer-Objekt mit Hashwert oder null, wenn ein Fehler auftritt.
   */
  private Integer bestimmeHashCode(final Object xml, final JAXBContext context) {
    try {
      Marshaller marshaller = context.createMarshaller();
      marshaller.setProperty(Marshaller.JAXB_FORMATTED_OUTPUT, new Boolean(true));
      marshaller.setProperty(Marshaller.JAXB_ENCODING, CHAR_ENCODING);

      ByteArrayOutputStream ba = new ByteArrayOutputStream();

      marshaller.marshal(new JAXBElement<Object>(new QName(DATEX2_NAMESPACE, DATEX2_ROOTTAG), Object.class, xml), ba);

      String xmlString = ba.toString();
      
      Integer hashCode = xmlString.hashCode();

      return hashCode;

    } catch (JAXBException ex) {
      return null;
    }
  }

  /**
   * TO BE REMOVED.
   * @return NO RETURNS
   */
  
  public String getDateAsString() {
	   DateFormat formatter = new SimpleDateFormat("yyyyMMdd");
	   return formatter.format(new Date());
	}

  @Override
  public void aktualisiere() {
    try {
      DatenSammlungsMethode dataMethod = plugin.getDatenSammlungsMethode();

      if (dataMethod.isSendeBereit()) {
        Map<ObjekteAtgAsp, Collection<ResultData> > data = dataMethod.ermittleDaten();
        Object d2l = plugin.getDatexKonvertierungDynamisch().getDatexII(data);

        if (d2l == null) {
          throw new Exception("Leeres XML-Dokument als Ergebnis.");
        }
        
        for (PublizierungsMethode m : plugin.getDynamischePublizierungsMethoden()) {
          m.publiziere(d2l, plugin.getDatexKonvertierungDynamisch().getJAXBContext(), null);
          LOGGER.info(logPrefix + "Publizierung " + m.getPublizierungsMethodenName() + " abgeschlossen.");
        }

        dataMethod.wurdePubliziert();
        plugin.getErweiterterLogger().sendeStatus("Dynamische Daten publiziert.");
      }

    } catch (Exception ex) {
      LOGGER.error(logPrefix + "Publikation dynamischer Daten fehlgeschlagen.", ex);
      plugin.getErweiterterLogger().sendeAnOperator(logPrefix 
          + "Publikation dynamischer Daten fehlgeschlagen: " + ex.getMessage());
    }
  }
}
