/**
 * SWE KEx-DatexII (de.bsvrz.kex.datex2)
 * 
 * Copyright (C) 2014 TraffGo Road GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * TraffGo Road GmbH
 * Ostwall 11-19
 * 47798 Krefeld
 * Phone: +49-2151-3295058
 * mailto: info@traffgoroad.de
 */

package de.bsvrz.kex.datex2.plugins.empfang.impl;

import java.io.ByteArrayOutputStream;
import java.io.InputStream;
import java.io.StringReader;
import java.net.URL;
import java.net.URLConnection;
import java.util.concurrent.CancellationException;
import java.util.zip.GZIPInputStream;

import javax.xml.bind.Unmarshaller;
import javax.xml.transform.Source;
import javax.xml.transform.stream.StreamSource;

import de.bsvrz.kex.datex2.plugins.definition.PluginImport;
import de.bsvrz.kex.datex2.plugins.empfang.EmpfangsMethode;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Klasse zum Abrufen von Daten vom MDM, Dekompression und Unmarshalling der XML-Struktur.
 *
 * @author TraffGo Road GmbH
 * @version $Id$
 */

public class MDMEmpfang implements EmpfangsMethode {
  /** Ausgabe der ersten Zeichen des empfangenen Streams. */
  private static final int MAX_RECV_BUFFER_EXTRACT = 1000;

  /** Timeout (Verbindungsaufbau und Datenbertragung) in Millisekunden. */
  private static final int TIMEOUT = 10000;
  
  /** Standard-Beschrnkung des Datenempfangs (in Bytes). */
  private static final int MAX_SIZE_DEFAULT = 20000000;

  /** Block-Gre beim Datenempfang (in Bytes). */
  private static final int CHUNK_SIZE = 1024;

  /** Referenz auf Plugin-Assembly-Klasse. */
  private PluginImport plugin;
  
  /** URL von der Daten abgerufen werden. */
  private String srcURL;
  
  /** Beschrnkung des Datenempfangs (in Bytes). */
  private Integer maxSize;
  
  /** Prfix fr Logging-Ausgaben. */
  private String logPrefix = "";

  /** Logger fr Debugausgaben. */
  private static final Debug LOGGER = Debug.getLogger();

  /**
   * Konstruktor. Das bergebene Objekt vom Typ KEx2Plugin ist bei Instantiierung
   * noch nicht vollstndig initialisiert.
   *
   * @param plugin
   *            Referenz auf Plugin-Assembly-Klasse
   */
  public MDMEmpfang(final PluginImport plugin) {
    this.plugin = plugin;
  }

  /** {@inheritDoc} */
  @Override
  public boolean initialisiere() {
    srcURL = plugin.getKonfigurationsModul().getString("URL");
    maxSize = plugin.getKonfigurationsModul().getInteger("MaximaleGroesse");
    if (maxSize == null) {
      maxSize = MAX_SIZE_DEFAULT;
    }
    logPrefix = "In Kontext '" + plugin.getKonfigurationsModul() + "': ";
    return true;
  }
  
  /** {@inheritDoc} */
  @Override
  public Object empfangeDaten() {
    InputStream in = null;
    ByteArrayOutputStream ba = null;
    
    try {
      URL url = new URL(srcURL);
      URLConnection connection = url.openConnection();
      connection.setConnectTimeout(TIMEOUT);
      connection.setReadTimeout(TIMEOUT);
      
      in = new GZIPInputStream(connection.getInputStream());

      int totalSize = 0;
      int readBytes;
      byte[] buf = new byte[CHUNK_SIZE];
      ba = new ByteArrayOutputStream();
      
      // Lese Daten in Chunks, Abbruch wenn berlauf

      while ((readBytes = in.read(buf)) > 0) {
        ba.write(buf, 0, readBytes);
        totalSize += readBytes;

        if (totalSize >= maxSize) {
          throw new CancellationException(logPrefix 
              + "Es wurden mehr als " + totalSize + " Bytes empfangen, Import wird abgebrochen.");
        }
      }

      LOGGER.info(logPrefix + "Datenimport vom MDM: " + totalSize + " Bytes");

      Source xmls = new StreamSource(new StringReader(ba.toString()));
     
      Unmarshaller m = (Unmarshaller) plugin.getDaVKonvertierungDynamisch().getJAXBContext().createUnmarshaller();
      Object result = m.unmarshal(xmls);
      if (result == null) {
        int len = ba.toString().length();
        if (len > MAX_RECV_BUFFER_EXTRACT) {
          len = MAX_RECV_BUFFER_EXTRACT;
        }
        LOGGER.info(logPrefix + "Datenimport nicht erfolgreich; Auszug der empfangenen Daten:"
          + ba.toString().substring(0, len));
      }
      return result;
    } catch (Exception ex) {
      LOGGER.error(logPrefix + "Datenbertragung vom MDM konnte nicht durchgefhrt werden.", ex);
      plugin.getErweiterterLogger().sendeAnOperator(logPrefix 
          + "Datenimport vom MDM konnte nicht durchgefhrt werden " + ex.getMessage());
      return null;
    } finally {
      if (in != null) {
        try { in.close(); } catch (Exception dummy) { in = null; }
      }
      if (ba != null) {
        try { ba.close(); } catch (Exception dummy) { ba = null; }
      }
    }
  }
}
