/**
 * SWE KEx-DatexII (de.bsvrz.kex.datex2)
 * 
 * Copyright (C) 2014 TraffGo Road GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * TraffGo Road GmbH
 * Ostwall 11-19
 * 47798 Krefeld
 * Phone: +49-2151-3295058
 * mailto: info@traffgoroad.de
 */

package de.bsvrz.kex.datex2.plugins.konvertierung.impl;

import java.util.Calendar;
import java.util.Collection;
import java.util.Date;
import java.util.Map;

import javax.naming.ConfigurationException;
import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.datatype.XMLGregorianCalendar;

import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.kex.datex2.allgemein.ObjekteAtgAsp;
import de.bsvrz.kex.datex2.plugins.definition.PluginExport;
import eu.datex2.schema.dynamisch.fahrstreifen.verkehrslage.ConfidentialityValueEnum;
import eu.datex2.schema.dynamisch.fahrstreifen.verkehrslage.CountryEnum;
import eu.datex2.schema.dynamisch.fahrstreifen.verkehrslage.D2LogicalModel;
import eu.datex2.schema.dynamisch.fahrstreifen.verkehrslage.ElaboratedData;
import eu.datex2.schema.dynamisch.fahrstreifen.verkehrslage.ElaboratedDataPublication;
import eu.datex2.schema.dynamisch.fahrstreifen.verkehrslage.Exchange;
import eu.datex2.schema.dynamisch.fahrstreifen.verkehrslage.HeaderInformation;
import eu.datex2.schema.dynamisch.fahrstreifen.verkehrslage.InformationStatusEnum;
import eu.datex2.schema.dynamisch.fahrstreifen.verkehrslage.InternationalIdentifier;
import eu.datex2.schema.dynamisch.fahrstreifen.verkehrslage.LocationByReference;
import eu.datex2.schema.dynamisch.fahrstreifen.verkehrslage.OverallPeriod;
import eu.datex2.schema.dynamisch.fahrstreifen.verkehrslage.PayloadPublication;
import eu.datex2.schema.dynamisch.fahrstreifen.verkehrslage.PredefinedLocationVersionedReference;
import eu.datex2.schema.dynamisch.fahrstreifen.verkehrslage.PredefinedNonOrderedLocationGroupVersionedReference;
import eu.datex2.schema.dynamisch.fahrstreifen.verkehrslage.ReferenceSettings;
import eu.datex2.schema.dynamisch.fahrstreifen.verkehrslage.TrafficStatus;
import eu.datex2.schema.dynamisch.fahrstreifen.verkehrslage.TrafficStatusEnum;
import eu.datex2.schema.dynamisch.fahrstreifen.verkehrslage.TrafficStatusValue;
import eu.datex2.schema.dynamisch.fahrstreifen.verkehrslage.Validity;
import eu.datex2.schema.dynamisch.fahrstreifen.verkehrslage.ValidityStatusEnum;

/**
 * Konvertierung Verkehrslage in das DATEX-II Format.
 * 
 * @author TraffGo Road GmbH
 * @version $Id$
 */

public class FSVerkehrslage extends StandardKonvertierungDynamisch {
  /** Faktor zur Umrechnung Sekunde => Millisekunde. */
    private static final int FACTOR_MILLI_TO_SECOND = 1000;
    
  /** Faktor zur Umrechnung [0;1] => Prozent [0;100]. */
    private static final int FACTOR_PERCENT = 100;

  /** Akzeptierte Datenverteiler-Objekttypen. */
  private static final String TYP_FAHRSTREIFEN = "typ.fahrStreifen";
  
  /** Akzeptierte Datenverteiler-Attributgruppen. */
  private static final String ATG_FAHRSTREIFEN = "atg.strfallZustand";
  
  /**
   * Konstruktor.
   * 
   * @param plugin Referenz auf Plugin-Definitions-Klasse.
   */
	  
	  public FSVerkehrslage(final PluginExport plugin) {
	    this.plugin = plugin;
	  }
	  
	  /***************************************************************************
	   * IN DEM FOLGENDEN ABSCHNITT WERDEN STRUKTURELL GLEICHE METHODEN GESAMMELT.
	   * 
	   * Diese Methoden werden in den einzelnen Plugins wiederholt - allerdings
	   * beziehen sich die JAXB-Elemente trotz gleichen Namens dann auf
	   * andere Klassen; eine nominale Gleichheit existiert also nicht.
	   ***************************************************************************
	   */

	  /**
	   * Gebe den in dieser Klasse verwendeten JAXBContext zurck. 
	   * 
	   * @throws JAXBException {@see javax.xml.bind.JAXBContext#newInstance(String)}
	   * @return JAXBContext Verwendeter Kontext
	   */

	  @Override
	  public JAXBContext getJAXBContext() throws JAXBException {
	    if (jaxbContext == null) {
	      jaxbContext = JAXBContext.newInstance(D2LogicalModel.class.getPackage().getName());
	    }
	    return jaxbContext;
	  }

	  /**
	   * Setzt die allgemeinen Angaben zur Publikation (Exchange, Language, PublicationCreator).
	   * 
	   * @param d2l D2LogicalModel
	   * @param payload Verwendeter Payload
	   */

	  private void setzeNationalIdentifier(final D2LogicalModel d2l, final PayloadPublication payload) {
	    InternationalIdentifier internationalIdentifier = new InternationalIdentifier();
	    internationalIdentifier.setCountry(CountryEnum.DE);
	    internationalIdentifier.setNationalIdentifier(nationalIdentifier);
	    
	    // Setze Exchange
	    
	    Exchange exchange = new Exchange();
	    exchange.setSupplierIdentification(internationalIdentifier);
	    d2l.setExchange(exchange);

	    // Setze Language
	    payload.setLang("DE");

	    payload.setPublicationCreator(internationalIdentifier);
	  }
	  
	  /***********************************************************************
	   * ENDE DES ABSCHNITTS STRUKTURELL GLEICHER METHODEN
	   ***********************************************************************
	   */

	  /** {@inheritDoc} */
	  @Override
	  public Object getDatexII(final Map<ObjekteAtgAsp, Collection<ResultData>> data) {
      String predefinedLocationVersion = holePredefinedLocationVersion();
	    D2LogicalModel d2l = new D2LogicalModel();
	    ElaboratedDataPublication payload = new ElaboratedDataPublication();
	    
	    setzeNationalIdentifier(d2l, payload);
	    
	    HeaderInformation headerInformation = new HeaderInformation();
	    headerInformation.setConfidentiality(ConfidentialityValueEnum.NO_RESTRICTION);
	    headerInformation.setInformationStatus(InformationStatusEnum.REAL);
	    payload.setLang("DE");

	    ReferenceSettings referenceSettings = new ReferenceSettings();
	    PredefinedNonOrderedLocationGroupVersionedReference predefinedNonOrderedLocationGroupVersionedReference = 
	        new PredefinedNonOrderedLocationGroupVersionedReference();
	    predefinedNonOrderedLocationGroupVersionedReference.setId(locationGroupId);
	    predefinedNonOrderedLocationGroupVersionedReference.setVersion(predefinedLocationVersion);
	    referenceSettings.setPredefinedNonOrderedLocationGroupReference(predefinedNonOrderedLocationGroupVersionedReference);
      referenceSettings.setTrafficStatusDefault(TrafficStatusEnum.FREE_FLOW);
	    
      XMLGregorianCalendar xmlGregorianCalendar = WerkzeugSammlung.konvertiereDatum(new Date());

      payload.setPublicationTime(xmlGregorianCalendar);
      payload.setHeaderInformation(headerInformation);
      payload.setReferenceSettings(referenceSettings);

	    for (ObjekteAtgAsp objekteAtgAsp : data.keySet()) {
	      if (!objekteAtgAsp.getTyp().getPid().equals(TYP_FAHRSTREIFEN) 
	          || (!objekteAtgAsp.getAttributGruppe().getPid().equals(ATG_FAHRSTREIFEN))) {
	        LOGGER.warning("Unpassende Konfiguration des Plugins: " 
	            + "Typ/Attributgruppe passt nicht zur dynamischen Konvertierung"); 
	      } else {
	        for (ResultData resultData : data.get(objekteAtgAsp)) {
	          try { 
	            LocationByReference locationByReference = new LocationByReference();
	            PredefinedLocationVersionedReference predefinedLocationVersionedReference = 
	                new PredefinedLocationVersionedReference();
	            predefinedLocationVersionedReference.setId(resultData.getObject().getPid());
	            predefinedLocationVersionedReference.setVersion(predefinedLocationVersion);
	            locationByReference.setPredefinedLocationReference(predefinedLocationVersionedReference);
	            ElaboratedData elaboratedData = new ElaboratedData();
	            TrafficStatus trafficStatus = new TrafficStatus();
	            if (resultData.hasData()) {
	              boolean hasElaboratedData = false;
	              String situation = resultData.getData().getTextValue("Situation").getValueText();
	              long t =  resultData.getData().getTimeValue("T").getMillis();
                float period = t / FACTOR_MILLI_TO_SECOND;

	              long horizont =  resultData.getData().getTimeValue("Horizont").getMillis();
                int horizontMilliSekunden = WerkzeugSammlung.getIntValue(horizont);

                float gueteIndex = WerkzeugSammlung.getFloatValue(resultData, "Gte", "Index", SMALL_MINUS_FLOAT, 2);
                Float gueteProzent = null;
                if (gueteIndex >= 0) {
                  gueteProzent = Float.valueOf((1.0f - gueteIndex) * FACTOR_PERCENT);
                }              

	              trafficStatus.setPertinentLocation(locationByReference);
	              trafficStatus.setMeasurementOrCalculationTime(xmlGregorianCalendar);
	              trafficStatus.setMeasurementOrCalculationPeriod(period);

                TrafficStatusValue trafficStatusValue = new TrafficStatusValue();
                trafficStatus.setTrafficStatus(trafficStatusValue);

                trafficStatusValue.setAccuracy(gueteProzent);

	              switch (situation.trim().toLowerCase()) {
	              case "strung":
	                trafficStatusValue.setDataError(true);
	                trafficStatusValue.setTrafficStatusValue(TrafficStatusEnum.UNKNOWN);
	                hasElaboratedData = false; // true
	                break;
	              case "keine aussage":
	                trafficStatusValue.setDataError(true);
	                trafficStatusValue.setTrafficStatusValue(TrafficStatusEnum.UNKNOWN);
	                hasElaboratedData = false; // true
	                break;
	              case "lebhafter verkehr":
	              case "dichter verkehr":
	                trafficStatusValue.setTrafficStatusValue(TrafficStatusEnum.HEAVY);
	                hasElaboratedData = true;
	                break;
	              case "zhflieender verkehr":
	              case "stockender verkehr":
	                trafficStatusValue.setTrafficStatusValue(TrafficStatusEnum.CONGESTED);
	                hasElaboratedData = true;
	                break;
	              case "stau":
	                trafficStatusValue.setTrafficStatusValue(TrafficStatusEnum.IMPOSSIBLE);
	                elaboratedData.setBasicData(trafficStatus);
	                hasElaboratedData = true;
	                break;
	               case "freier verkehr":
	                  if (payload.getElaboratedDatas().isEmpty()) {
	                    trafficStatusValue.setTrafficStatusValue(TrafficStatusEnum.FREE_FLOW);
	                    hasElaboratedData = true;
	                  } else {
	                    //Defaultwert Freier Verkehr = TrafficStatusEnum.FREE_FLOW
	                    hasElaboratedData = false;
	                  }
	                  break;
	              default:
	                trafficStatusValue.setDataError(true);
                  trafficStatusValue.setTrafficStatusValue(TrafficStatusEnum.UNKNOWN);                    
                  hasElaboratedData = false;  // true
                  break;
	              }
	              if (hasElaboratedData) {
	                elaboratedData.setBasicData(trafficStatus);
	                if (horizont > 0) { 
	                  elaboratedData.setValidity(setzeValidity(horizontMilliSekunden));
	                  elaboratedData.setForecast(true);
	                }
	                payload.getElaboratedDatas().add(elaboratedData);
	              } // hasElaboratedData
	            } 
	          } catch (Exception ex) {
              LOGGER.fine(logPrefix + "Verwerfe einen Eintrag wegen Fehler in der Verarbeitung", ex);
            }
	        } // resultDate == null
	      } // end for
	    }
      d2l.setPayloadPublication(payload);
      if (payload.getElaboratedDatas().isEmpty()) {
        LOGGER.warning(logPrefix + "Kein valider Datensatz verfgbar, Publikation wird abgebrochen");
        return null;
      } else { 
        return d2l;
      }
	  }

	  /**
	   * Setzt ein Start und Enddatum in ein DatexII Validity-Objekt. 
	   * 
	   * @param horizont
	   *          Integer Millisekunden als Intervalldauer
	   * @return Validity
	   */
	
	  protected Validity setzeValidity(final int horizont) {
		  
		  Validity validity = new Validity();
		  validity.setValidityStatus(ValidityStatusEnum.DEFINED_BY_VALIDITY_TIME_SPEC);
		  OverallPeriod overallPeriod = new OverallPeriod();
		  Date datumStart = new Date();
		  Calendar kalender = Calendar.getInstance();
		  kalender.setTime(datumStart);
		  kalender.add(Calendar.MILLISECOND, horizont);
		  Date datumEnd = kalender.getTime();
		  
		  XMLGregorianCalendar xmlGregorianCalendarStart = WerkzeugSammlung.konvertiereDatum(datumStart);
		  XMLGregorianCalendar xmlGregorianCalendarEnd = WerkzeugSammlung.konvertiereDatum(datumEnd);
		  
		  overallPeriod.setOverallStartTime(xmlGregorianCalendarStart);
		  overallPeriod.setOverallEndTime(xmlGregorianCalendarEnd);
		  validity.setValidityTimeSpecification(overallPeriod);
		  return validity;	    
	  }
	
  @Override
  public void initialisiere(final Collection<ObjekteAtgAsp> objects) throws ConfigurationException {
    initialisiereBasisFelder();
    
    initialisiereStandardIntervall();
  }
}
